# coding: utf-8

import io
import sys
import struct
import base64

if sys.version_info[0] == 2:
	import collections as abc
	_integer_types = (int, long)
else:
	import collections.abc as abc
	_integer_types = (int, )
_list_types = (abc.Sequence, abc.Set)

_str_type = type(u"")
_bytes_type = type(b"")

from .data import Tagging, Undefined

class EncoderError(Exception):
	pass

class Mapping(object):
	__slots__ = ('map')
	def __init__(self, map):
		self.map = map
def mapping(obj):
	if isinstance(obj, abc.Mapping):
		return obj
	else:
		return Mapping(obj)

class Encoder(object):
	def __init__(self, output):
		self.output = output

	def encode(self, val):
		if isinstance(val, _bytes_type):
			self.encode_bytestring(val)
		elif isinstance(val, _str_type):
			self.encode_textstring(val)
		elif isinstance(val, float):
			self.encode_float(val)
		elif isinstance(val, bool):
			self.encode_boolean(val)
		elif isinstance(val, _integer_types):
			self.encode_integer(val)
		elif isinstance(val, abc.Sequence) or isinstance(val, abc.Set):
			self.encode_list(val)
		elif isinstance(val, abc.Mapping):
			self.encode_dict(val)
		elif isinstance(val, Tagging):
			self.encode_tagging(val)
		elif val is Undefined:
			self.encode_undefined()
		elif val is None:
			self.encode_null()
		elif isinstance(val, Mapping):
			val = val.map
			if isinstance(val, abc.Sized):
			    self.encode_dict(val)
			else:
			    self.encode_infinite_dict(val)
		elif isinstance(val, abc.Iterable):
			self.encode_infinite_list(val)
		else:
			raise EncoderError("val of type {} is not serializable".format(type(val)))

	def encode_list(self, list):
		self._write(_encode_ibyte(4, len(list)))
		for elem in list:
			self.encode(elem)

	def encode_dict(self, dict):
		self._write(_encode_ibyte(5, len(dict)))
		for key, value in dict.items():
			self.encode(key)
			self.encode(value)

	def encode_bytestring(self, bytestring):
		self._write(_encode_ibyte(2, len(bytestring)))
		self._write(bytestring)

	def encode_textstring(self, textstring):
		string_ = textstring.encode("utf-8")
		self._write(_encode_ibyte(3, len(string_)))
		self._write(string_)

	def encode_float(self, float):
		self._write(b"\xfb")
		self._write(struct.pack(">d", float))

	def encode_integer(self, integer):
		if integer < 0:
			integer = -integer - 1
			try:
				self._write(_encode_ibyte(1, integer))
			except TypeError:
				raise EncoderError("Encoding integers lower than -18446744073709551616 is not supported")
		else:
			try:
				self._write(_encode_ibyte(0, integer))
			except TypeError:
				raise EncoderError("Encoding integers larger than 18446744073709551615 is not supported")

	def encode_tagging(self, tagging):
		try:
			self._write(_encode_ibyte(6, tagging.tag))
		except TypeError:
			raise EncoderError("Encoding tag larger than 18446744073709551615 is not supported")
		self.encode(tagging.obj)

	def encode_boolean(self, boolean):
		if boolean is True:
			self._write(_encode_ibyte(7, 21))
		elif boolean is False:
			self._write(_encode_ibyte(7, 20))

	def encode_null(self):
		self._write(_encode_ibyte(7, 22))

	def encode_undefined(self):
		self._write(b"\xf7")

	def encode_infinite_textstring(self, iterable):
		self._write(b"\x7f")
		for elem in iterable:
			if not isinstance(elem, _str_type):
				raise EncoderError("Object of type {} is not valid in infinite textstring".format(type(elem)))
			self.encode(elem)
		self._write(b"\xff")

	def encode_infinite_bytestring(self, iterable):
		self._write(b"\x5f")
		for elem in iterable:
			if not isinstance(elem, _bytes_type):
				raise EncoderError("Object of type {} is not valid in infinite bytestring".format(type(elem)))
			self.encode(elem)
		self._write(b"\xff")

	def encode_infinite_list(self, iterable):
		self._write(b"\x9f")
		for elem in iterable:
			self.encode(elem)
		self._write(b"\xff")

	def encode_infinite_dict(self, iterable):
		self._write(b"\xbf")
		for key, value in iterable:
			self.encode(key)
			self.encode(value)
		self._write(b"\xff")

	def _write(self, val):
		self.output.write(val)

class InfiniteEncoder(Encoder):
	chunksize = 8

	def __init__(self, output, chunksize=None):
		Encoder.__init__(self, output)
		if chunksize is not None:
			self.chunksize = chunksize

	def encode_list(self, object):
		self.encode_infinite_list(object)

	def encode_dict(self, object):
		self.encode_infinite_dict(object.items())

	def encode_textstring(self, object):
		if len(object) <= self.chunksize:
			Encoder.encode_textstring(self, object)
		else:
			chunks = int((len(object) - 1) / self.chunksize) + 1
			generator = (object[n*self.chunksize:(n+1)*self.chunksize] for n in range(chunks))
			self.encode_infinite_textstring(generator)

	def encode_bytestring(self, object):
		if len(object) <= self.chunksize:
			Encoder.encode_bytestring(self, object)
		else:
			chunks = int((len(object) - 1) / self.chunksize) + 1
			generator = (object[n*self.chunksize:(n+1)*self.chunksize] for n in range(chunks))
			self.encode_infinite_bytestring(generator)

def _encode_ibyte(major, length):
	if length < 24:
		return struct.pack('>B', (major << 5) | length)
	elif length < 256:
		return struct.pack('>BB', (major << 5) | 24, length)
	elif length < 65536:
		return struct.pack('>BH', (major << 5) | 25, length)
	elif length < 4294967296:
		return struct.pack('>BI', (major << 5) | 26, length)
	elif length < 18446744073709551616:
		return struct.pack('>BQ', (major << 5) | 27, length)
	else:
		return None

def dump(obj, io, cls=Encoder, *args, **kwargs):
	cls(io, *args, **kwargs).encode(obj)

def dumps(obj, *args, **kwargs):
	buf = io.BytesIO()
	dump(obj, buf, *args, **kwargs)
	return buf.getvalue()

def dumph(*args, **kwargs):
	return base64.b16encode(dumps(*args, **kwargs)).decode("utf-8")

__all__ = ["Encoder", "InfiniteEncoder", "EncoderError", "dump", "dumps", "dumph", "mapping"]




# coding: utf-8

import collections

class Tagging(object):
        __slots__ = ("tag", "obj")
        def __init__(self, tag, obj):
            self.tag = tag
            self.obj = obj
        def __eq__(self, other):
            return isinstance(other, Tagging) and self.tag == other.tag and self.obj == other.obj
        def __hash__(self):
            return hash((self.tag, self.obj))

class _Undefined(object):
	_instance = None

	def __new__(cls, *args, **kwargs):
		if not isinstance(cls._instance, cls):
			cls._instance = object.__new__(cls, *args, **kwargs)
		return cls._instance

	def __str__(self):
		return "Undefined"

	def __repr__(self):
		return "Undefined"

Undefined = _Undefined()



# coding: utf-8

import io
import struct
import math
import sys
import base64

from .data import Tagging, Undefined

class InvalidCborError(Exception):
	pass

class _Break(InvalidCborError):
	def __init__(self):
		InvalidCborError.__init__(self, "Invalid BREAK code occurred")

class Decoder(object):
	def __init__(self, input):
		self._jump_table = [
			lambda *args: self.decode_integer(*args, sign=False),
			lambda *args: self.decode_integer(*args, sign=True),
			self.decode_bytestring,
			self.decode_textstring,
			self.decode_list,
			self.decode_dict,
			self.decode_tagging,
			self.decode_other
		]
		self.input = input

	def decode(self):
		mtype, ainfo = self._decode_ibyte()
		try:
			decoder = self._jump_table[mtype]
		except KeyError:
			raise InvalidCborError("Invalid major type {}".format(mtype))
		return decoder(mtype, ainfo)

	def decode_integer(self, mtype, ainfo, sign=False):
		res = self._decode_length(ainfo)
		if sign is True:
			return -1 - res
		else:
			return res

	def decode_bytestring(self, mtype, ainfo):
		length = self._decode_length(ainfo)
		if length is None:
			res = bytearray()
			while True:
				mtype_, ainfo_ = self._decode_ibyte()
				if (mtype_, ainfo_) == (7, 31):
					break
				if mtype_ != 2:
					pass
				res.extend(self.decode_bytestring(mtype_, ainfo_))
			return bytes(res)
		else:
			return self._read(length)

	def decode_textstring(self, mtype, ainfo):
		length = self._decode_length(ainfo)
		if length is None:
			res = bytearray()
			while True:
				mtype_, ainfo_ = self._decode_ibyte()
				if (mtype_, ainfo_) == (7, 31):
					break
				if mtype_ != 3:
					pass
				res.extend(self.decode_bytestring(mtype_, ainfo_))
			return res.decode("utf-8")
		else:
			return self._read(length).decode("utf-8")

	def decode_list(self, mtype, ainfo):
		length = self._decode_length(ainfo)
		if length is None:
			res = []
			while True:
				try:
					res.append(self.decode())
				except _Break:
					break
			return res
		else:
			res = [None for _ in range(length)]
			for n in range(length):
				res[n] = self.decode()
			return res

	def decode_dict(self, mtype, ainfo):
		length = self._decode_length(ainfo)
		if length is None:
			res = {}
			try:
				while True:
					key = self.decode()
					value = self.decode()
					res[key] = value
			except _Break:
				pass
			return res
		else:
			res = {}
			for n in range(length):
				key, value = self.decode(), self.decode()
				res[key] = value
			return res

	def decode_tagging(self, mtype, ainfo):
		length = self._decode_length(ainfo)
		return Tagging(length, self.decode())

	def decode_half_float(self, mtype, ainfo):
		half = struct.unpack(">H", self._read(2))[0]
		valu = (half & 0x7fff) << 13 | (half & 0x8000) << 16
		if ((half & 0x7c00) != 0x7c00):
			return math.ldexp(struct.unpack("!f", struct.pack("!I", valu))[0], 112)
		return struct.unpack("!f", struct.pack("!I", valu | 0x7f800000))[0]

	def decode_single_float(self, mtype, ainfo):
		return struct.unpack(">f", self._read(4))[0]

	def decode_double_float(self, mtype, ainfo):
		return struct.unpack(">d", self._read(8))[0]

	def decode_other(self, mtype, ainfo):
		if ainfo == 20:
			return False
		elif ainfo == 21:
			return True
		elif ainfo == 22:
			return None
		elif ainfo == 23:
			return Undefined
		elif ainfo == 25:
			return self.decode_half_float(mtype, ainfo)
		elif ainfo == 26:
			return self.decode_single_float(mtype, ainfo)
		elif ainfo == 27:
			return self.decode_double_float(mtype, ainfo)
		elif ainfo == 31:
			raise _Break()

	def _decode_ibyte(self):
		byte = self._read(1)[0]
		if isinstance(byte, str):
			byte = ord(byte)
		return (byte & 0b11100000) >> 5, byte & 0b00011111

	def _decode_length(self, ainfo):
		if ainfo < 24:
			return ainfo
		elif ainfo == 24:
			return from_bytes(self._read(1))
		elif ainfo == 25:
			return from_bytes(self._read(2))
		elif ainfo == 26:
			return from_bytes(self._read(4))
		elif ainfo == 27:
			return from_bytes(self._read(8))
		elif ainfo == 31:
			return None
		raise InvalidCborError("Invalid additional information {}".format(ainfo))

	def _read(self, n):
		m = self.input.read(n)
		if len(m) != n:
			raise InvalidCborError("Expected {} bytes, got {} bytes instead".format(n, len(m)))
		return m

class StandardDecoder(Decoder):
	def __init__(self, input, tagging_hooks=None):
		Decoder.__init__(self, input)
		if tagging_hooks is None:
			tagging_hooks = {}
		self.tagging_hooks = tagging_hooks
	
	def decode_tagging(self, mtype, ainfo):
		tagged = Decoder.decode_tagging(self, mtype, ainfo)
		if tagged.tag in self.tagging_hooks:
			return self.tagging_hooks[tagged.tag](tagged.tag, tagged.object)
		else:
			return tagged
	
	def register_tagging(self, tag_id, hook):
		self.tagging_hooks[tag_id] = hook

	def unregister_tagging(self, tag_id):
		del self.tagging_hooks[tag_id]
	
def load(fp, cls=Decoder, *args, **kwargs):
	return cls(fp, *args, **kwargs).decode()

def loads(data, cls=Decoder, *args, **kwargs):
	return cls(io.BytesIO(data), *args, **kwargs).decode()

def loadh(data, *args, **kwargs):
	return loads(base64.b16decode(data), *args, **kwargs)

__all__ = ["InvalidCborError", "Decoder", "load", "loads", "loadh"]

if sys.version_info[0] == 2:
    import struct
    from_formats = (None, '>B', '>H', None, '>I', None, None, None, '>Q')

    def from_bytes(val):
        return struct.unpack(from_formats[len(val)], val)[0]
else:
    def from_bytes(val):
        return int.from_bytes(val, "big")
	



# coding: utf-8

from . import decoder
from . import encoder
from . import data

__all__ = [
	"decoder",
	"encoder",
	"dump",
	"dumps",
	"dumph",
	"load",
	"loads",
	"loadh",
	"Tagging",
	"Undefined"
]

dump = encoder.dump
dumps = encoder.dumps
dumph = encoder.dumph
mapping = encoder.mapping

load = decoder.load
loads = decoder.loads
loadh = decoder.loadh

Tagging = data.Tagging
Undefined = data.Undefined

